const Web3Utils = require('web3-utils');
const { isSignerAssociated } = require('./trust-clues');
const { AIRLINE_SEGMENT_ID, HOTEL_SEGMENT_ID } = require('../constants');

const { config } = require('../config');

let _entrypoint;

function getInstance () {
  return config.wtLibs;
}

function _setEntrypointAddress (address) {
  config.entrypointAddress = address;
  _entrypoint = null;
}

function getEntrypoint () {
  if (_entrypoint) {
    return _entrypoint;
  }
  return getInstance().getEntrypoint(config.entrypointAddress);
}

async function getHotelDirectory () {
  return getEntrypoint().getSegmentDirectory(HOTEL_SEGMENT_ID);
}

async function getAirlineDirectory () {
  return getEntrypoint().getSegmentDirectory(AIRLINE_SEGMENT_ID);
}

async function getFactory () {
  return getEntrypoint().getOrganizationFactory();
}

function getTrustClueClient () {
  return getInstance().getTrustClueClient();
}

async function checkGuarantee (trustClueClient, guarantee, hotelAddress) {
  const data = JSON.parse(Web3Utils.hexToUtf8(guarantee.claim));
  await trustClueClient.verifySignedData(guarantee.claim, guarantee.signature, async (actualSigner) => {
    if (hotelAddress !== data.hotel) {
      throw new Error(`Guarantee seems to be for a different hotel at ${data.hotel}.`);
    }
    if (!data.expiresAt) {
      throw new Error('Guarantee has no expiration.');
    }
    if ((new Date()).getTime() > data.expiresAt) {
      throw new Error(`Guarantee expired at ${new Date(data.expiresAt)}.`);
    }
    // TODO guarantor might be an account or org.id
    /* if (Web3Utils.toChecksumAddress(data.guarantor) !== actualSigner) {
      throw new Error(`Expected signer '${data.guarantor}' does not match the recovered one '${actualSigner}'`);
    } */
    if (!await isSignerAssociated(actualSigner, data.guarantor)) {
      throw new Error(`Signer (${actualSigner}) is not associated with guarantor (${data.guarantor}).`);
    }
  });
}

async function passesTrustworthinessTest (hotelAddress, guarantee) {
  if (!config.checkTrustClues) {
    return true;
  }
  const trustClueClient = getInstance().getTrustClueClient();
  try {
    // 1. check guarantee
    if (guarantee && guarantee.claim && guarantee.signature) {
      await checkGuarantee(trustClueClient, guarantee, hotelAddress);
    }
    // 2. check org.id trust levels
    const trustworthinessTestResults = await trustClueClient.interpretAllValues(hotelAddress);
    // 3. return single boolean value - this expects all clues to have interpret function that returns boolean
    return trustworthinessTestResults
      .map((v) => {
        if (!v.value) {
          console.warn(`Organization at ${hotelAddress} does not pass ${v.name}`);
        }
        return v.value === true;
      })
      .indexOf(false) === -1;
  } catch (e) {
    config.logger.warn(`Cannot establish trust level for '${hotelAddress}': ${e.toString()}`);
    return false;
  }
}

module.exports = {
  getInstance,
  getFactory,
  getHotelDirectory,
  getAirlineDirectory,
  getTrustClueClient,
  passesTrustworthinessTest,
  _setEntrypointAddress,
};
